@@ -33,6 +33,10 @@
  */
 
 #include <netinet/in.h>
+#include <sys/types.h>
+
+#ifndef __dnscap_dnscap_common_h
+#define __dnscap_dnscap_common_h
 
 /*
  * setup MY_BPFTIMEVAL as the timeval structure that bpf packets
@@ -82,3 +86,5 @@ typedef void output_t(const char *descr,
 
 #define DIR_INITIATE	0x0001
 #define DIR_RESPONSE	0x0002
+
+#endif /* __dnscap_dnscap_common_h */
